import asyncio
import datetime
import logging
import os

from aiohttp import web
from aiohttp_swagger import setup_swagger

from calorify_users.baseapp import DatabaseWebApp
from calorify_users.middleware import jwt_token_verification
from calorify_users.tasks.emailverification import (
    send_unsent_verification_emails,
)
from calorify_users.view.authenticate import (
    authenticate,
    confirm_token,
)
from calorify_users.view.emailverification import (
    confirm_reset_password,
    reset_password,
    verify_email_token,
)
from calorify_users.view.role import (
    create_role,
    get_roles,
    remove_role,
)
from calorify_users.view.user import (
    add_user,
    get_single_user,
    get_users,
    remove_single_user,
    update_single_user,
)
from calorify_users.config import config


def generate_server_secret():
    return os.urandom(512//8)


async def run_periodically(loop, period):
    if type(period) == datetime.timedelta:
        period = period.seconds

    try:
        while True:
            loop.create_task(
                send_unsent_verification_emails(app["db_session"]))
            await asyncio.sleep(period)
    except asyncio.CancelledError:
        logging.info("Job runner done.")


def create_users_app(database_url):
    app = DatabaseWebApp(
        app_name="calorify-users", app_config=config,
        middlewares=[jwt_token_verification],
        database_url=database_url)
    app["server_secret"] = generate_server_secret()

    app.router.add_route('POST', "/authenticate", authenticate)
    app.router.add_route('POST', "/confirm-token", confirm_token)

    app.router.add_route('GET', "/user", get_users)
    app.router.add_route('POST', "/user", add_user)
    app.router.add_route('GET', "/user/{user_id}", get_single_user)
    app.router.add_route('PATCH', "/user/{user_id}", update_single_user)
    app.router.add_route('DELETE', "/user/{user_id}", remove_single_user)

    app.router.add_route('GET', "/verify-email", verify_email_token)
    app.router.add_route('GET', "/password-reset", reset_password)
    app.router.add_route('GET', "/confirm-password-reset",
                         confirm_reset_password)

    app.router.add_route('GET', "/role", get_roles)
    app.router.add_route('POST', "/role", create_role)
    app.router.add_route('DELETE', "/role/{name}", remove_role)

    # not supported in pip-packaged version of aiohttp-swagger.
    # security_definitions = {
    #    "type": "http",
    #    "scheme": "bearer",
    #    "bearerFormat": "JWT",
    #}
    setup_swagger(app)
    return app


async def on_shutdown(app):
    app["tasks"].cancel()


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    app = create_users_app(config.get("database_url"))
    app.on_shutdown.append(on_shutdown)

    app["tasks"] = loop.create_task(
        run_periodically(loop, datetime.timedelta(minutes=30)))
    web.run_app(app, host="127.0.0.1", port=5000)
