import logging
import sys

from aiohttp import web
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from calorify_users.middleware import database_rollback_on_error


class BaseDatabaseApp(web.Application):
    """
    Connect to `database_url` and store the session on self["db_session"].
    """
    def __init__(self, *args, **kwargs):
        try:
            database_url = kwargs.pop("database_url")
        except KeyError:
            raise Exception(
                "BaseDatabaseApp requires a `database_url` keyword argument.")

        middlewares = [database_rollback_on_error]
        middlewares.extend(kwargs.get("middlewares", []))
        kwargs["middlewares"] = middlewares
        super(BaseDatabaseApp, self).__init__(*args, **kwargs)

        engine = create_engine(database_url)
        self["db_session"] = Session(bind=engine)


class DatabaseWebApp(BaseDatabaseApp):

    def __init__(self, *args, **kwargs):
        app_name = kwargs.pop("app_name", "db-webapp")
        config = kwargs.pop("app_config", {})
        logger = kwargs.pop("logger", None)
        logger = self._configure_logger(logger, app_name)

        super(DatabaseWebApp, self).__init__(*args, logger=logger, **kwargs)

        self._app_config = config
        self._app_name = app_name
        self.on_response_prepare.append(self._cors_on_prepare)

        self.router.add_route(
            "OPTIONS", "/{tail:.*}", self._cors_options_handler)

    def _configure_logger(self, logger, app_name):
        if logger is None:
            for logger_name in [None, "aiohttp.access", app_name]:
                logger = logging.getLogger(logger_name)
                logger.setLevel(logging.INFO)
                logger.addHandler(logging.StreamHandler(sys.stderr))
        return logger

    async def _cors_on_prepare(self, request, response):
        response.headers.add(
            "Access-Control-Allow-Origin",
            self._app_config.get("allow_origin", "*"))
        response.headers.add("Access-Control-Allow-Methods",
                             "GET, POST, OPTIONS, PATCH, DELETE")
        for allowed_header in ("accept", "access-control-allow-origin",
                               "content-type", "authorization",
                               "access-control-allow-methods"):
            response.headers.add(
                "Access-Control-Allow-Headers", allowed_header)

    @staticmethod
    async def _cors_options_handler(request):
        return web.Response(text="")
