import jwt
from aiohttp import web
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import joinedload

from calorify_users.model.user import User
from calorify_users.view.generic import error_response


def verify_token(token, request):
    """Return True if token is valid, raise an exception otherwise."""
    decoded_token = jwt.decode(token, request.app["server_secret"])
    dbsession = request.app["db_session"]
    user = dbsession.query(User).options(joinedload(User.roles)).filter(
        User.email == decoded_token["email"]).one_or_none()
    assert user is not None, (
        "Somehow ended up with a verified token and no user to match it. "
        "Has the user been removed?")
    request["authorized_user"] = user
    request["authorized_roles"] = [role.name for role in user.roles]
    assert request["authorized_roles"] == decoded_token["roles"], (
        "Your roles are out of date, please re-authenticate.")
    return True


@web.middleware
async def jwt_token_verification(request, handler):
    """Verifies any Authorization bearer tokens.

    On successful verification, sets request["authorized_roles"] and
    request["authorized_user"] to be used by handlers.
    """
    auth_header = request.headers.get("AUTHORIZATION", "").strip()
    if auth_header[:7].lower() == "bearer ":
        token = auth_header.split(" ", 1)[1].strip()
        try:
            verify_token(token, request)
        except AssertionError as error:
            return error_response(request, str(error), {}, status=401)
        except jwt.exceptions.PyJWTError:
            return error_response(request, "Invalid token.", {}, status=401)
    response = await handler(request)
    return response


@web.middleware
async def database_rollback_on_error(request, handler):
    try:
        response = await handler(request)
        return response
    except SQLAlchemyError:
        request.app["db_session"].rollback()
        raise
