import base64
import datetime
import os

from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    String,
)
from sqlalchemy.orm import (
    relationship,
    validates,
)

from calorify_users.helpers import validate_email
from calorify_users.model import Base


def datetime_24h_from_now():
    """Return a datetime 24h from now."""
    return datetime.datetime.now() + datetime.timedelta(hours=24)


class EmailVerification(Base):
    __tablename__ = "email_verification"

    user_id = Column(Integer(), ForeignKey("user.id"), primary_key=True)
    user = relationship(
        "User", back_populates="email_verification", uselist=False)
    email = Column(String(), primary_key=True, nullable=False)
    token = Column(String(32), nullable=False)
    expires = Column(DateTime(), default=datetime_24h_from_now)
    url_callback = Column(String(), nullable=True)
    sent = Column(Boolean(), default=False, nullable=False)
    is_email_change = Column(Boolean(), default=False, nullable=False)
    is_password_reset = Column(Boolean(), default=False, nullable=False)

    @validates("email")
    def email_validator(self, key, value):
        return validate_email(value)


def create_or_update_email_verification(
        db_session, user_id, email, url_callback=None, is_email_change=False,
        is_password_reset=False):
    """Creates or updates an EmailVerification with a new token."""

    assert not(is_email_change and is_password_reset)

    # Generate a new 32-byte (24*4/3) long verification token string.
    token_value = base64.b64encode(os.urandom(24)).decode("ascii")

    email_token = db_session.query(
        EmailVerification).filter(
            EmailVerification.user_id == user_id,
            EmailVerification.email == email).one_or_none()
    if email_token is None:
        email_token = EmailVerification(
            user_id=user_id, email=email, token=token_value,
            url_callback=url_callback, is_email_change=is_email_change,
            is_password_reset=is_password_reset)
        db_session.add(email_token)
    else:
        email_token.token = token_value
        email_token.expires = datetime_24h_from_now()
    return email_token
