import datetime
import pytest
import sqlalchemy.exc

from calorify_users.model.emailverification import (
    EmailVerification,
    create_or_update_email_verification,
)
from calorify_users.model.user import User


class TestEmailVerification:

    def test_new_record(self, testdb_session):
        """
        A new EmailVerification record referencing a user can be added.
        """
        new_user = User(name="Woo Hoo", password="pwd")
        testdb_session.add(new_user)
        testdb_session.flush()

        verification = EmailVerification(
            user_id=new_user.id, email="woo@hoo", token="foo")
        testdb_session.add(verification)
        testdb_session.flush()

        assert verification.user_id == new_user.id, (
            "Verification user_id not set correctly.")
        assert verification.user == new_user, (
            "Verification user not set correctly.")
        assert verification.email == "woo@hoo", (
            "Verification email address not set correctly.")
        assert verification.token == "foo", (
            "Verification token not set correctly.")

    def test_urlcallback_is_supported(self, testdb_session):
        """
        A new EmailVerification record can have an url_callback.
        """
        new_user = User(name="Woo Hoo", password="pwd")
        testdb_session.add(new_user)
        testdb_session.flush()

        verification = EmailVerification(
            user_id=new_user.id, email="woo@hoo", token="foo",
            url_callback="http://somewhere/trusted?welcome")
        testdb_session.add(verification)
        testdb_session.flush()

        assert verification.url_callback == "http://somewhere/trusted?welcome"

    def test_is_email_change_flag(self, testdb_session):
        """
        An EmailVerification record can have an `is_email_change` boolean flag.
        """
        new_user = User(name="Woo Hoo", password="pwd")
        testdb_session.add(new_user)
        testdb_session.flush()

        verification = EmailVerification(
            user_id=new_user.id, email="woo@hoo", token="foo",
            is_email_change=True)
        testdb_session.add(verification)
        testdb_session.flush()

        assert verification.is_email_change

    def test_sent_flag(self, testdb_session):
        """
        An EmailVerification record can have a `sent` boolean flag.
        """
        new_user = User(name="Woo Hoo", password="pwd")
        testdb_session.add(new_user)
        testdb_session.flush()

        verification = EmailVerification(
            user_id=new_user.id, email="woo@hoo", token="foo", sent=True)
        testdb_session.add(verification)
        testdb_session.flush()

        assert verification.sent

    def test_requires_user(self, testdb_session):
        """
        A new EmailVerification must be linked to an existing user.
        """
        with pytest.raises(sqlalchemy.exc.IntegrityError):
            verification = EmailVerification(
                user_id=1, email="woo@hoo", token="foo")
            testdb_session.add(verification)
            testdb_session.flush()

    def test_requires_token(self, testdb_session):
        """
        A new EmailVerification must be linked to an existing user.
        """
        new_user = User(name="Woo Hoo", password="pwd")
        testdb_session.add(new_user)
        testdb_session.flush()
        with pytest.raises(sqlalchemy.exc.IntegrityError):
            verification = EmailVerification(
                user_id=new_user.id, email="woo@hoo")
            testdb_session.add(verification)
            testdb_session.flush()

    def test_default_expires(self, testdb_session):
        """
        A new EmailVerification record gets a default expires 24h from now.
        """
        new_user = User(name="Woo Hoo", password="pwd")
        testdb_session.add(new_user)
        testdb_session.flush()

        # Allow expires to be 24h from now ±1s.
        expected_expires = (
            datetime.datetime.now() + datetime.timedelta(hours=24))
        expires_from = expected_expires - datetime.timedelta(seconds=1)
        expires_to = expected_expires + datetime.timedelta(seconds=1)

        verification = EmailVerification(
            user_id=new_user.id, email="woo@hoo", token="foo")
        testdb_session.add(verification)
        testdb_session.flush()

        assert verification.expires is not None, (
            "EmailVerification.expires should be set to 24h in the future.")
        assert expires_from < verification.expires < expires_to, (
            "EmailVerification.expires does not match expected 24h delay.")

    def test_duplicate_emails(self, testdb_session):
        """
        Different users can attempt to validate the same email address.

        Eg. if someone typoes their address, the actual owner should not be
        stopped from registering.
        """
        user1 = User(name="Woo Hoo", password="pwd")
        user2 = User(name="Hoo Woo", password="pwd")
        testdb_session.add(user1)
        testdb_session.add(user2)
        testdb_session.flush()

        verification1 = EmailVerification(
            user_id=user1.id, email="woo@hoo", token="foo")
        verification2 = EmailVerification(
            user_id=user2.id, email="woo@hoo", token="foo")
        testdb_session.add(verification1)
        testdb_session.add(verification2)
        testdb_session.flush()

    def test_does_email_validation(self, testdb_session):
        """Ensure email validation is done on the email field."""
        at_sign_missing = "E-mail address must contain an '@' character."
        new_user = User(name="Woo Hoo", password="pwd")
        testdb_session.add(new_user)
        testdb_session.flush()
        with pytest.raises(AssertionError) as error:
            EmailVerification(
                user_id=new_user.id, email="woo.hoo", token="foo")
        assert str(error.value) == at_sign_missing, (
            "Failed for a wrong reason.")


class TestCreateOrUpdateEmailVerification:

    def test_create_new(self, testdb_session):
        user = User(name="foo", password="pwd")
        testdb_session.add(user)
        testdb_session.flush()

        # Allow expires to be 24h from now ±1s.
        expected_expires = (
            datetime.datetime.now() + datetime.timedelta(hours=24))
        expires_from = expected_expires - datetime.timedelta(seconds=1)
        expires_to = expected_expires + datetime.timedelta(seconds=1)

        verification = create_or_update_email_verification(
            testdb_session, user.id, "woo@hoo",
            url_callback="http://somewhere/trusted", is_email_change=True)
        testdb_session.flush()
        assert verification.user_id == user.id, (
            "Verification user_id not set correctly.")
        assert verification.email == "woo@hoo", (
            "Verification email address not set correctly.")
        assert len(verification.token) == 32, (
            "Verification token not generated correctly or missing.")
        assert expires_from < verification.expires < expires_to, (
            "EmailVerification.expires does not match expected 24h delay.")
        assert verification.url_callback == "http://somewhere/trusted"
        assert verification.is_email_change

    def test_update_existing(self, testdb_session):
        user = User(name="foo", password="pwd")
        testdb_session.add(user)
        testdb_session.flush()
        verification = create_or_update_email_verification(
            testdb_session, user.id, "woo@hoo")
        # Patch the expiry to be 12h from now to ensure this gets updated.
        verification.expires = (
            datetime.datetime.now() + datetime.timedelta(hours=12))
        testdb_session.flush()
        old_token = verification.token

        # Allow updated expires to be 24h from now ±1s.
        expected_expires = (
            datetime.datetime.now() + datetime.timedelta(hours=24))
        expires_from = expected_expires - datetime.timedelta(seconds=1)
        expires_to = expected_expires + datetime.timedelta(seconds=1)

        updated_verification = create_or_update_email_verification(
            testdb_session, user.id, "woo@hoo")
        testdb_session.flush()
        assert len(updated_verification.token) == 32, (
            "Verification token not generated correctly or missing.")
        assert updated_verification.token != old_token, (
            "Verification token was not re-generated.")
        assert expires_from < updated_verification.expires < expires_to, (
            "EmailVerification.expires was not updated to another 24h delay.")
