import aiosmtplib
import asyncio
import datetime
import logging
import os.path

from email.mime.text import MIMEText
from sqlalchemy.sql.expression import false
from urllib.parse import (
    urlencode,
    urljoin,
)
from calorify_users.model.emailverification import EmailVerification


async def send_unsent_verification_emails(db_session, config):
    logging.info("Running periodic job to send unsent verification emails.")

    # We rely on server timezone remaining the same, but it's not too bad
    # for email verifications to be valid for a couple more or fewer hours
    # if server indeed moves timezones.
    now = datetime.datetime.now()

    # XXX: We should lock the verification table rows for update while we are
    # sending an email to avoid a different job runner picking up the same
    # verifications to send out, resulting in duplicate emails.
    outstanding_verifications = db_session.query(EmailVerification).filter(
        EmailVerification.sent == false(),
        EmailVerification.expires >= now).all()  # noqa: E712
    if len(outstanding_verifications) == 0:
        logging.info("Nothing to do.")

    # This will open a single email connection for every email verification
    # that needs to be sent to allow for parallel execution.
    # We could switch to a pool of fixed number of connections and then batch
    # send emails over them, but not worth the effort since we should have
    # more than a few only if the sending keeps failing (we email one out as
    # soon as it's created).
    tasks = [send_email_verification(verification, db_session, config)
             for verification in outstanding_verifications]
    return await asyncio.gather(*tasks)


async def send_email_verification(verification, db_session, config):
    template_dir_path = os.path.normpath(
        os.path.join(os.path.dirname(__file__), "..", "templates"))

    verification_type = None
    if verification.is_email_change:
        template_file = "user-email-change-verification.txt"
        verification_type = "email change"
    elif verification.is_password_reset:
        template_file = "password-reset-verification.txt"
        verification_type = "password reset"
    else:
        template_file = "new-user-email-verification.txt"
        verification_type = "new user email"
    template_path = os.path.join(template_dir_path, template_file)
    with open(template_path) as template_file:
        template_subject = template_file.readline().strip()
        # Ignore second line.
        template_file.readline()
        template_text = template_file.read()

    text = template_text.format(
        user_name=verification.user.name,
        product_name=config["product_name"],
        verification_url=urljoin(
            config["base_url"], "verify-email?" +
            urlencode({"token": verification.token})),
        url_callback=urljoin(
            verification.url_callback,
            urlencode({"token": verification.token})))

    logging.info("Sending {} verification for {}.".format(
        verification_type, verification.email))
    email_config = config.get("email", {})

    message = MIMEText(text)
    message["from"] = email_config["from"]
    message["to"] = "{} <{}>".format(
        verification.user.name, verification.email)
    message["subject"] = template_subject

    server_config = email_config.get("server", {})
    smtp = aiosmtplib.SMTP(hostname=server_config.get("host"),
                           port=server_config.get("port"),
                           use_tls=server_config.get("use_tls", False))
    await smtp.connect()
    if server_config.get("use_starttls") and not server_config.get("use_tls"):
        # We only issue STARTTLS if use_starttls is True but use_tls is not.
        await smtp.starttls()
    await smtp.send_message(message)
    verification.sent = True
    db_session.commit()
