import datetime
import pytest

import calorify_users.tasks.emailverification

from calorify_users.model.emailverification import EmailVerification
from calorify_users.model.user import User
from calorify_users.tasks.emailverification import (
    send_email_verification,
    send_unsent_verification_emails,
)


class MockSMTPServer:
    instances = []

    def __init__(self, hostname=None, port=None, use_tls=False):
        self.hostname = hostname
        self.port = port
        self.use_tls = False
        self.use_starttls = False
        self.calls = []
        MockSMTPServer.instances.append(self)

    async def starttls(self, *args, **kwargs):
        self.calls.append(("starttls", args, kwargs))
        self.use_starttls = True

    async def connect(self, *args, **kwargs):
        self.calls.append(("connect", args, kwargs))

    async def send_message(self, *args, **kwargs):
        self.calls.append(("send_message", args, kwargs))


@pytest.fixture
def mock_smtp_servers(request):
    MockSMTPServer.instances = []
    real_SMTP_class = calorify_users.tasks.emailverification.aiosmtplib.SMTP

    def restore_SMTP_class():
        calorify_users.tasks.emailverification.aiosmtplib.SMTP = (
            real_SMTP_class)

    calorify_users.tasks.emailverification.aiosmtplib.SMTP = MockSMTPServer
    request.addfinalizer(restore_SMTP_class)
    return MockSMTPServer.instances


@pytest.fixture
def sample_verification(testdb_session):
    """An EmailVerification record attached to a User."""
    working_token = "0123456789abcdef0123456789abcdef"
    user = User(name="Woo Hoo", password="password")
    testdb_session.add(user)
    testdb_session.flush()
    verification = EmailVerification(
        user_id=user.id, email="woo@hoo", token=working_token)
    testdb_session.add(verification)
    testdb_session.commit()
    return verification


class TestSendEmailVerification:
    sample_config = {
        "product_name": "Super Thing",
        "base_url": "http://super-thing",
        "email": {
            "from": "danilo@localhost",
            "server": {
                "host": "localhost",
                "port": None,  # Use default (25 or 465 for TLS)
                "use_tls": False,
                "use_starttls": False,
            },
        },
    }

    async def test_server_config(
            self, testdb_session, mock_smtp_servers, sample_verification):
        """Hostname, port and use_tls are taken from the config."""
        await send_email_verification(
            sample_verification, testdb_session, self.sample_config)

        smtp_server = mock_smtp_servers[0]
        assert (
            smtp_server.hostname == "localhost" and
            smtp_server.port is None and
            smtp_server.use_tls is False)
        call_names = [call[0] for call in smtp_server.calls]
        assert call_names == ["connect", "send_message"]

    async def test_server_starttls(
            self, testdb_session, mock_smtp_servers, sample_verification):
        """If starttls is used, it is issued after the connect call."""
        server_config = self.sample_config.copy()
        server_config["email"]["server"]["use_starttls"] = True
        await send_email_verification(sample_verification, testdb_session,
                                      server_config)

        smtp_server = mock_smtp_servers[0]
        call_names = [call[0] for call in smtp_server.calls]
        assert call_names[:2] == ["connect", "starttls"]

    async def test_server_no_starttls_with_use_tls(
            self, testdb_session, mock_smtp_servers, sample_verification):
        """If TLS is used, starttls is not even if use_starttls is set."""
        sample_config = self.sample_config.copy()
        server_config = sample_config["email"]["server"]
        server_config["use_tls"], server_config["use_starttls"] = True, True

        await send_email_verification(sample_verification, testdb_session,
                                      sample_config)

        smtp_server = mock_smtp_servers[0]
        call_names = [call[0] for call in smtp_server.calls]
        assert "starttls" not in call_names

    async def test_message_from(
            self, testdb_session, mock_smtp_servers, sample_verification):
        """Message is sent from an address in config."""
        await send_email_verification(sample_verification, testdb_session,
                                      self.sample_config)

        smtp_server = mock_smtp_servers[0]
        send_message_call = [call for call in smtp_server.calls
                             if call[0] == "send_message"][0]
        message = send_message_call[1][0]
        assert message["from"] == self.sample_config["email"]["from"]

    async def test_message_subject(
            self, testdb_session, mock_smtp_servers, sample_verification):
        """Subject is a pre-set value."""
        await send_email_verification(sample_verification, testdb_session,
                                      self.sample_config)

        smtp_server = mock_smtp_servers[0]
        send_message_call = [call for call in smtp_server.calls
                             if call[0] == "send_message"][0]
        message = send_message_call[1][0]
        assert message["subject"] == "Calorify Email Verification"

    async def test_message_to(
            self, testdb_session, mock_smtp_servers, sample_verification):
        """Message is sent to an address from verification including a name."""
        await send_email_verification(sample_verification, testdb_session,
                                      self.sample_config)

        smtp_server = mock_smtp_servers[0]
        send_message_call = [call for call in smtp_server.calls
                             if call[0] == "send_message"][0]
        message = send_message_call[1][0]
        assert message["to"] == "{} <{}>".format(
            sample_verification.user.name, sample_verification.email)

    async def test_link_in_message(
            self, testdb_session, mock_smtp_servers, sample_verification):
        """Link containing the token is inside the mail content."""
        working_token = sample_verification.token

        await send_email_verification(sample_verification, testdb_session,
                                      self.sample_config)

        smtp_server = mock_smtp_servers[0]
        send_message_call = [call for call in smtp_server.calls
                             if call[0] == "send_message"][0]
        message = send_message_call[1][0]
        assert "?token=" + working_token in message.get_payload()

    async def test_uses_new_user_template(
            self, testdb_session, mock_smtp_servers, sample_verification):
        """
        If verification is not for an email change, use new user template.
        """
        sample_verification.is_email_change = False
        sample_verification.is_password_reset = False
        testdb_session.flush()

        await send_email_verification(sample_verification, testdb_session,
                                      self.sample_config)

        smtp_server = mock_smtp_servers[0]
        send_message_call = [call for call in smtp_server.calls
                             if call[0] == "send_message"][0]
        message = send_message_call[1][0]
        assert "register" in message.get_payload()

    async def test_uses_email_change_template(
            self, testdb_session, mock_smtp_servers, sample_verification):
        """
        If verification is for an email change, use appropriate template.
        """
        sample_verification.is_email_change = True
        testdb_session.flush()

        await send_email_verification(sample_verification, testdb_session,
                                      self.sample_config)

        smtp_server = mock_smtp_servers[0]
        send_message_call = [call for call in smtp_server.calls
                             if call[0] == "send_message"][0]
        message = send_message_call[1][0]
        assert "change your email" in message.get_payload()

    async def test_uses_password_reset_template(
            self, testdb_session, mock_smtp_servers, sample_verification):
        """
        If verification is for a password reset, use a different template.
        """
        sample_verification.is_password_reset = True
        sample_verification.url_callback = "http://reset-password-here"
        testdb_session.flush()

        await send_email_verification(sample_verification, testdb_session,
                                      self.sample_config)

        smtp_server = mock_smtp_servers[0]
        send_message_call = [call for call in smtp_server.calls
                             if call[0] == "send_message"][0]
        message = send_message_call[1][0]

        payload = message.get_payload()
        assert "reset your password" in payload
        assert "http://reset-password-here/token=" in payload


@pytest.fixture
def two_sample_verifications(testdb_session):
    """An EmailVerification record attached to a User."""
    user1 = User(name="Woo Hoo", password="password")
    user2 = User(name="Foo Bar", password="password")
    testdb_session.add_all([user1, user2])
    testdb_session.flush()
    working_token1 = "0123456789abcdef0123456789abcdef"
    working_token2 = "fedcba9876543210fedcba9876543210"
    verification1 = EmailVerification(
        user_id=user1.id, email="woo@hoo", token=working_token1)
    verification2 = EmailVerification(
        user_id=user2.id, email="foo@bar", token=working_token2)
    testdb_session.add_all([verification1, verification2])
    testdb_session.commit()
    return verification1, verification2


class TestSendUnsentVerificationEmails:
    sample_config = {
        "product_name": "Super Thing",
        "base_url": "http://super-thing",
        "email": {
            "from": "danilo@localhost",
            "server": {
                "host": "localhost",
                "port": None,  # Use default (25 or 465 for TLS)
                "use_tls": False,
                "use_starttls": False,
            },
        },
    }

    async def test_none(self, testdb_session, mock_smtp_servers):
        """No verifications means no email messages."""
        await send_unsent_verification_emails(
            testdb_session, self.sample_config)

        assert len(mock_smtp_servers) == 0

    async def test_one(self, testdb_session, mock_smtp_servers,
                       sample_verification):
        """One verification triggers a single send message."""
        await send_unsent_verification_emails(
            testdb_session, self.sample_config)

        assert len(mock_smtp_servers) == 1
        smtp_server = mock_smtp_servers[0]
        call_names = [call[0] for call in smtp_server.calls]
        assert call_names == ["connect", "send_message"]

    async def test_expired(self, testdb_session, mock_smtp_servers,
                           sample_verification):
        """Expired verification is not sent."""
        sample_verification.expires -= datetime.timedelta(hours=36)
        testdb_session.flush()

        await send_unsent_verification_emails(
            testdb_session, self.sample_config)

        assert len(mock_smtp_servers) == 0

    async def test_multiple(self, testdb_session, mock_smtp_servers,
                            two_sample_verifications):
        """
        Two unsent verifications trigger two SMTP connections for each message.
        """
        verification1, verification2 = two_sample_verifications

        await send_unsent_verification_emails(
            testdb_session, self.sample_config)

        assert len(mock_smtp_servers) == 2
        smtp_server1 = mock_smtp_servers[0]
        smtp_server2 = mock_smtp_servers[1]
        send_message_call1 = [call for call in smtp_server1.calls
                              if call[0] == "send_message"][0]
        send_message_call2 = [call for call in smtp_server2.calls
                              if call[0] == "send_message"][0]
        message1 = send_message_call1[1][0]

        message2 = send_message_call2[1][0]
        verification1_from = "{} <{}>".format(
            verification1.user.name, verification1.email)
        verification2_from = "{} <{}>".format(
            verification2.user.name, verification2.email)

        assert (
            (message1["to"] == verification1_from and
             message2["to"] == verification2_from and
             verification1.token in message1.get_payload() and
             verification2.token in message2.get_payload()) or
            (message2["to"] == verification1_from and
             message1["to"] == verification2_from and
             verification2.token in message1.get_payload() and
             verification1.token in message2.get_payload()))
