import aiohttp.web
import json
import pytest
import types
from aiohttp.test_utils import make_mocked_request

from calorify_users.middleware import jwt_token_verification
from calorify_users.model.user import User
from calorify_users.model.userrole import UserRole
from calorify_users.model.role import Role
from calorify_users.view.authenticate import (
    create_jwt_token,
)


async def sample_handler(request):
    return aiohttp.web.json_response(
        {"roles": request.get("authorized_roles"),
         "user": request.get("authorized_user").email})


@pytest.fixture
def api_client(loop, aiohttp_client, testdb_session):
    app = aiohttp.web.Application(middlewares=[jwt_token_verification])
    app["server_secret"] = "secret"
    app["db_session"] = testdb_session

    app.router.add_get('/test', sample_handler)
    yield loop.run_until_complete(aiohttp_client(app))


class TestJWTTokenVerification:
    async def test_wraps_async_request_handler(self):
        async def sample_handler(request):
            return {"handler_called": True}

        request = make_mocked_request('GET', '/', headers={})
        response = await jwt_token_verification(request, sample_handler)
        assert response == {"handler_called": True}

    async def test_ignores_non_bearer_authorization_headers(self):
        async def sample_handler(request):
            return {"roles": request.get("authorized_roles"),
                    "user": request.get("authorized_user")}

        request = make_mocked_request(
            'GET', '/', headers={"Authorization": "basic foo@bar"})
        response = await jwt_token_verification(request, sample_handler)
        assert response == {"roles": None, "user": None}

    async def test_ignores_bad_bearer_header(self):
        async def sample_handler(request):
            return {"roles": request.get("authorized_roles"),
                    "user": request.get("authorized_user")}

        request = make_mocked_request(
            'GET', '/', headers={"Authorization": "bearer "})
        response = await jwt_token_verification(request, sample_handler)
        assert response == {"roles": None, "user": None}

    async def test_bad_jwt_bearer_token(self):
        # This should be an async def, but it shouldn't ever be executed,
        # so we keep it a simple variable to satisfy python-coverage and not
        # trigger flake8.
        sample_handler = None

        request = make_mocked_request(
            'GET', '/', app=aiohttp.web.Application(),
            headers={"Authorization": "bearer not-a-jwt-token"})
        request.app["server_secret"] = "secret"
        response = await jwt_token_verification(request, sample_handler)
        assert response.status == 401
        data = json.loads(response.text)
        assert data == {"message": "Invalid token.", "hints": {}}

    async def test_failed_verification(self, testdb_session):
        # This should be an async def, but it shouldn't ever be executed,
        # so we keep it a simple variable to satisfy python-coverage and not
        # trigger flake8.
        sample_handler = None

        user = types.SimpleNamespace(
            id=37, name="Woo Hoo", email="woo@hoo",
            roles=[types.SimpleNamespace(name="admin"),
                   types.SimpleNamespace(name="user")])
        token = create_jwt_token(user, "secret")

        request = make_mocked_request(
            'GET', '/', app=aiohttp.web.Application(),
            headers={"Authorization": "bearer {}".format(token),
                     "accept": "application/json"})
        request.app["server_secret"] = "secret"
        request.app["db_session"] = testdb_session

        response = await jwt_token_verification(request, sample_handler)
        assert response.status == 401
        data = json.loads(response.text)
        error_message = (
            "Somehow ended up with a verified token and no user to match it. "
            "Has the user been removed?")
        assert data == {"message": error_message, "hints": {}}

    async def test_successful_verification(self, testdb_session):
        async def sample_handler(request):
            return {"roles": request.get("authorized_roles"),
                    "user": request.get("authorized_user")}

        user = User(email="woo@hoo", name="Woo Hoo", password="pwd")
        role1 = Role(name="admin")
        role2 = Role(name="test")
        testdb_session.add_all([user, role1, role2])
        testdb_session.flush()
        testdb_session.add_all([UserRole(user_id=user.id, role_id=role1.id),
                                UserRole(user_id=user.id, role_id=role2.id)])
        testdb_session.flush()
        token = create_jwt_token(user, "secret")

        request = make_mocked_request(
            'GET', '/', app=aiohttp.web.Application(),
            headers={"Authorization": "bearer {}".format(token)})
        request.app["server_secret"] = "secret"
        request.app["db_session"] = testdb_session

        response = await jwt_token_verification(request, sample_handler)
        assert response == {"roles": ["admin", "test"], "user": user}

    async def test_verification_fails_when_roles_change(self, testdb_session):
        """
        If a role is added after the token has been issued, request
        gets the new role among authorized_roles as well.
        """
        sample_handler = None

        user = User(email="woo@hoo", name="Woo Hoo", password="pwd")
        role1 = Role(name="admin")
        testdb_session.add_all([user, role1])
        testdb_session.flush()
        testdb_session.add(UserRole(user_id=user.id, role_id=role1.id))
        testdb_session.flush()
        token = create_jwt_token(user, "secret")

        assert user.roles == [role1]

        role2 = Role(name="test")
        testdb_session.add(role2)
        testdb_session.flush()
        testdb_session.add(UserRole(user_id=user.id, role_id=role2.id))
        testdb_session.commit()

        request = make_mocked_request(
            'GET', '/', app=aiohttp.web.Application(),
            headers={"Authorization": "bearer {}".format(token)})
        request.app["server_secret"] = "secret"
        request.app["db_session"] = testdb_session

        response = await jwt_token_verification(request, sample_handler)
        assert response.status == 401
        data = json.loads(response.text)
        assert data == {
            "message": "Your roles are out of date, please re-authenticate.",
            "hints": {}}

    async def test_integration_with_app(self, api_client):
        user = User(email="woo@hoo", name="Woo Hoo", password="pwd")
        role = Role(name="admin")
        db_session = api_client.app["db_session"]
        db_session.add_all([user, role])
        db_session.flush()
        db_session.add(UserRole(user_id=user.id, role_id=role.id))
        db_session.flush()
        token = create_jwt_token(user, "secret")

        response = await api_client.get(
            "/test",
            headers={"Authorization": "bearer {}".format(token)})
        assert response.status == 200, "Failed call to verify JWT token."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == {"roles": ["admin"], "user": user.email}
