import aiohttp.web
import pytest

from calorify_users.middleware import jwt_token_verification
from calorify_users.model.user import User
from calorify_users.model.userrole import UserRole
from calorify_users.model.role import Role
from calorify_users.permissions import require_role
from calorify_users.view.authenticate import create_jwt_token


@require_role("admin")
async def admin_handler(request):
    return aiohttp.web.Response(text="admin success")


@require_role("admin", "manager")
async def manager_handler(request):
    return aiohttp.web.Response(text="manager success")


@pytest.fixture
def api_client(loop, aiohttp_client, testdb_session):
    app = aiohttp.web.Application(middlewares=[jwt_token_verification])
    app["server_secret"] = "secret"
    app["db_session"] = testdb_session

    app.router.add_get('/admin', admin_handler)
    app.router.add_get('/manager', manager_handler)
    yield loop.run_until_complete(aiohttp_client(app))


class TestRequireRole:

    async def test_successful_verification(self, api_client):
        user = User(email="woo@hoo", name="Woo Hoo", password="pwd")
        role = Role(name="admin")
        db_session = api_client.app["db_session"]
        db_session.add_all([user, role])
        db_session.flush()
        db_session.add(UserRole(user_id=user.id, role_id=role.id))
        db_session.commit()
        token = create_jwt_token(user, "secret")

        response = await api_client.get(
            "/admin",
            headers={"Authorization": "bearer {}".format(token)})
        assert response.status == 200, "Failed call to gather roles."
        assert response.content_type == "text/plain", (
            "Unexpected content type.")
        data = await response.text()
        assert data == "admin success"

    async def test_failed_verification(self, api_client):
        user = User(email="woo@hoo", name="Woo Hoo", password="pwd")
        role = Role(name="non-admin")
        db_session = api_client.app["db_session"]
        db_session.add_all([user, role])
        db_session.flush()
        db_session.add(UserRole(user_id=user.id, role_id=role.id))
        db_session.commit()
        token = create_jwt_token(user, "secret")

        response = await api_client.get(
            "/admin",
            headers={"Authorization": "bearer {}".format(token),
                     "Accept": "application/json"})
        assert response.status == 401, "Unexpected status code."
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Unauthorized", "hints": {}}

    async def test_multiple_roles(self, api_client):
        user = User(email="woo@hoo", name="Woo Hoo", password="pwd")
        role = Role(name="manager")
        db_session = api_client.app["db_session"]
        db_session.add_all([user, role])
        db_session.flush()
        db_session.add(UserRole(user_id=user.id, role_id=role.id))
        db_session.commit()
        token = create_jwt_token(user, "secret")

        response = await api_client.get(
            "/manager",
            headers={"Authorization": "bearer {}".format(token)})
        assert response.status == 200, "Failed call to gather roles."
        assert response.content_type == "text/plain", (
            "Unexpected content type.")
        data = await response.text()
        assert data == "manager success"
