import crypt
import jwt

from calorify_users.model.user import User
from calorify_users.view.generic import (
    adapt_response,
    error_response,
)


def create_jwt_token(user, secret, expires=None):
    payload = {
        "id": user.id,
        "name": user.name,
        "email": user.email,
        "roles": [role.name for role in user.roles],
    }
    if expires:
        payload["exp"] = expires
    return jwt.encode(payload, secret, algorithm='HS512').decode('ascii')


async def authenticate(request):
    """
    ---
    description: Log a user in and return a JWT token back.
    tags:
    - Users
    produces:
    - application/json
    responses:
        "200":
            description: return a JWT token containing user's roles and
                details.
    """
    user_data = await request.post()

    db_session = request.app["db_session"]
    server_secret = request.app["server_secret"]
    if user_data.get("email") and user_data.get("password"):
        # Attempt to authenticate this user.
        user = db_session.query(User).filter(
            User.email == user_data.get("email")).one_or_none()
        if user is not None:
            # Validate password by re-hashing it using the stored, hashed
            # password as the salt.
            hashed_password = crypt.crypt(
                user_data.get("password"), user.password_hash)
            password_valid = user.password_hash == hashed_password
            if password_valid:
                return adapt_response(
                    request, {"status": "logged in",
                              "token": create_jwt_token(user, server_secret)})

    return error_response(request, "Password invalid.", {}, status=403)


async def confirm_token(request):
    """
    ---
    description: Confirm JWT token has been issued and signed by this server.
    tags:
    - Users
    produces:
    - application/json
    responses:
        "200":
            description: token has been issued and signed by this server for
                a still valid user and roles.
        "401":
            description:
                User has been removed or their roles have changed,
                return "Invalid token." message.
    """
    if request.get("authorized_user") is not None:
        return adapt_response(request, {"result": "token-confirmed"})
    else:
        return error_response(request, "Invalid token.", {}, status=401)
