import datetime

from aiohttp import web
from sqlalchemy.sql.expression import true

from calorify_users.model.emailverification import (
    EmailVerification,
    create_or_update_email_verification,
)
from calorify_users.model.user import User
from calorify_users.tasks.emailverification import send_email_verification
from calorify_users.view.authenticate import create_jwt_token
from calorify_users.view.generic import (
    adapt_response,
    error_response,
)
from calorify_users.view.user import validate_url_callback


async def verify_email_token(request):
    """
    ---
    description:
        Verifies e-mail address by providing received token in the `token`
        field.
    tags:
    - Emails
    produces:
    - application/json
    responses:
        "200":
            description: return a new user record along with Link relation.
    """
    user_data = request.query
    db_session = request.app["db_session"]

    if len(user_data.get("token", "")) != 32:
        return error_response(request, "Invalid token.", {})

    verification = db_session.query(EmailVerification).filter(
        EmailVerification.token == user_data.get("token"),
        EmailVerification.expires > datetime.datetime.now()).one_or_none()
    if verification is None:
        return error_response(request, "Token not found or expired.", {})

    url_callback = verification.url_callback

    # This update and the following delete should always remain in the same
    # transaction.
    verification.user.email = verification.email

    # This also removes any potential outstanding verification issued for
    # another user for the same e-mail address (assuming a typo or malice).
    db_session.query(EmailVerification).filter(
        EmailVerification.email == verification.email).delete()
    db_session.commit()

    if url_callback:
        raise web.HTTPFound(location=url_callback)

    return adapt_response(request, {"user_id": verification.user_id})


async def reset_password(request):
    """
    ---
    description:
        Ask for a password reset through email confirmation.
        Provide a `url_callback` to append "token" to where user can
        reset their password.
    tags:
    - Password
    produces:
    - application/json
    responses:
        "200":
            description: Password reset email scheduled to be sent.
        "400":
            description:
                No user email to reset password for provided, or
                URL callback was invalid or disallowed.
    """
    user_data = await request.post()

    db_session = request.app["db_session"]
    if user_data.get("email"):
        user = db_session.query(User).filter(
            User.email == user_data.get("email")).one_or_none()
        if user is not None:
            error_messages = []
            hints = {"fields": {}}
            # Create a password reset email.
            url_callback = validate_url_callback(
                user_data.get("url_callback"), request.app._app_config,
                error_messages, hints)
            if not error_messages:
                verification = create_or_update_email_verification(
                    db_session, user.id, user.email,
                    url_callback=url_callback, is_password_reset=True)
                db_session.commit()
                request.loop.create_task(
                    send_email_verification(
                        verification, db_session, request.app._app_config))
            else:
                return error_response(request, "Invalid URL callback.", {})
        return adapt_response(request, {"status": "success"})
    else:
        return error_response(request, "Email not specified.", {})


async def confirm_reset_password(request):
    """
    ---
    description:
        Confirms password reset token and issues a new JWT token to
        use for password change.
    tags:
    - Users
    produces:
    - application/json
    responses:
        "200":
            description:
                return temporary JWT token to use to set password using
                "PATCH /user/{id}".
        "400":
            description: Invalid or expired token.
    """
    user_data = await request.post()
    db_session = request.app["db_session"]

    if len(user_data.get("token", "")) != 32:
        return error_response(request, "Invalid token.", {})

    verification = db_session.query(EmailVerification).filter(
        EmailVerification.token == user_data.get("token"),
        EmailVerification.expires > datetime.datetime.now(),
        EmailVerification.is_password_reset == true()).one_or_none()
    if verification is None:
        return error_response(request, "Token not found or expired.", {})

    server_secret = request.app["server_secret"]
    in_30_mins = datetime.datetime.utcnow() + datetime.timedelta(minutes=30)
    token = create_jwt_token(verification.user, server_secret,
                             expires=in_30_mins)

    db_session.query(EmailVerification).filter(
        EmailVerification.email == verification.email).delete()
    db_session.commit()

    return adapt_response(request, {"status": "success", "token": token})
