from calorify_users.model.role import Role
from calorify_users.permissions import require_role
from calorify_users.view.generic import (
    adapt_response,
    error_response,
)


@require_role("@admin", "@manager")
async def get_roles(request):
    """
    ---
    description:
        Lists all available roles.
    tags:
    - Roles
    produces:
    - application/json
    responses:
        "200":
            description: return a list of all available roles.
    """
    db_session = request.app["db_session"]
    role_data = [role.name for role in db_session.query(Role).all()]
    return adapt_response(request, {"roles": role_data})


@require_role("@admin")
async def create_role(request):
    """
    ---
    description:
        Create a role with name `name`.
    tags:
    - Roles
    produces:
    - application/json
    responses:
        "200":
            description: return created role.
    """
    db_session = request.app["db_session"]
    user_data = await request.post()
    if "name" in user_data and len(user_data["name"]) > 1:
        existing_role = db_session.query(Role).filter(
            Role.name == user_data["name"]).one_or_none()
        if existing_role is not None:
            return error_response(
                request,
                "Role '{}' already exists.".format(user_data["name"]),
                {})
        role = Role(name=user_data["name"])
    else:
        return error_response(
            request, "Role name must be at least 2 characters long.", {})
    db_session.add(role)
    return adapt_response(request, {"name": role.name})


@require_role("@admin")
async def remove_role(request):
    """
    ---
    description:
        Remove a role.
    tags:
    - Roles
    produces:
    - application/json
    responses:
        "200":
            description: return created role.
    """
    db_session = request.app["db_session"]
    role_name = request.match_info["name"]
    existing_role = db_session.query(Role).filter(
        Role.name == role_name).one_or_none()
    if existing_role is None:
        return error_response(
            request,
            "Role '{}' not found.".format(role_name),
            {}, status=404)
    db_session.delete(existing_role)
    return adapt_response(request, {"status": "ok"})
