import base64
import json
import jwt
import pytest
import types

from aiohttp import web

from calorify_users.middleware import jwt_token_verification
from calorify_users.model.user import User
from calorify_users.model.userrole import UserRole
from calorify_users.model.role import Role
from calorify_users.view.authenticate import (
    authenticate,
    confirm_token,
    create_jwt_token,
)


@pytest.fixture
def api_client(loop, aiohttp_client, testdb_session):
    app = web.Application()
    app["db_session"] = testdb_session
    app["server_secret"] = "secret"

    app.router.add_post('/test', authenticate)
    yield loop.run_until_complete(
        aiohttp_client(app, headers={"accept": "application/json"}))


class TestAuthenticate:
    async def test_success(self, api_client):
        db_session = api_client.app["db_session"]
        db_session.add(User(email="woo@hoo", name="Woo Hoo", password="pwd"))
        db_session.flush()

        response = await api_client.post(
            "/test", data={"email": "woo@hoo", "password": "pwd"})
        assert response.status == 200, "Failed authentication."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()

        expected_token = (
            "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzUxMiJ9."
            "eyJpZCI6MSwibmFtZSI6IldvbyBIb28iLCJlb"
            "WFpbCI6Indvb0Bob28iLCJyb2xlcyI6W119."
            "trPJkbch83mFhUTDvpubPIqmZJLZnVt_94hagzV44lo"
            "XvzRXiju8UbCP0URZvgZ4afaRVWf-t7KuIJnxcnwNkg")
        assert data == {"status": "logged in",
                        "token": expected_token}, "Unexpected JSON content."

    async def test_success_with_roles(self, api_client):
        db_session = api_client.app["db_session"]
        user = User(email="woo@hoo", name="Woo Hoo", password="pwd")
        role1 = Role(name="admin")
        role2 = Role(name="test")
        db_session.add_all([user, role1, role2])
        db_session.flush()
        db_session.add_all([UserRole(user_id=user.id, role_id=role1.id),
                            UserRole(user_id=user.id, role_id=role2.id)])

        response = await api_client.post(
            "/test", data={"email": "woo@hoo", "password": "pwd"})
        assert response.status == 200, "Failed authentication."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()

        token = data["token"]
        payload = jwt.decode(token, api_client.app["server_secret"])
        assert {"id": user.id,
                "name": "Woo Hoo",
                "email": "woo@hoo",
                "roles": ["admin", "test"]} == payload

    async def test_wrong_password(self, api_client):
        db_session = api_client.app["db_session"]
        db_session.add(User(email="woo@hoo", name="Woo Hoo", password="pwd"))
        db_session.flush()

        response = await api_client.post(
            "/test", data={"email": "woo@hoo", "password": "wrong-password"})
        assert response.status == 403, "Unexpected status code."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == {"message": "Password invalid.", "hints": {}}, (
            "Unexpected JSON content.")

    async def test_no_user(self, api_client):
        response = await api_client.post(
            "/test", data={"email": "woo@hoo", "password": "pwd"})
        assert response.status == 403, "Unexpected status code."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == {"message": "Password invalid.", "hints": {}}, (
            "Unexpected JSON content.")

    async def test_email_not_submitted(self, api_client):
        response = await api_client.post(
            "/test", data={"password": "pwd"})
        assert response.status == 403, "Unexpected status code."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == {"message": "Password invalid.", "hints": {}}, (
            "Unexpected JSON content.")

    async def test_password_not_submitted(self, api_client):
        response = await api_client.post(
            "/test", data={"email": "woo@hoo"})
        assert response.status == 403, "Unexpected status code."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == {"message": "Password invalid.", "hints": {}}, (
            "Unexpected JSON content.")


def create_auth_app(testdb_session):
    """
    Returns a web.Application instance using JWT token verification
    and routing /test URL to confirm_token view.
    """
    app = web.Application(middlewares=[jwt_token_verification])
    app["db_session"] = testdb_session
    app["server_secret"] = "secret"
    app.router.add_post('/test', confirm_token)
    return app


@pytest.fixture
def regular_api_client(loop, aiohttp_client, testdb_session):
    app = create_auth_app(testdb_session)
    user = User(email="woo@hoo", name="Woo Hoo", password="pwd")
    testdb_session.add(user)
    testdb_session.commit()
    token = create_jwt_token(user, app["server_secret"])
    headers = {"Authorization": "bearer {}".format(token),
               "Accept": "application/json"}
    app["test_authorized_user"] = user
    yield loop.run_until_complete(aiohttp_client(app, headers=headers))


class TestConfirmToken:
    async def test_token_accepted(self, regular_api_client):
        """
        Confirms token is stil valid.
        """
        response = await regular_api_client.post("/test")
        assert response.status == 200
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"result": "token-confirmed"}

    async def test_token_invalid_because_of_new_role(self, regular_api_client):
        """
        Confirms token is not valid anymore.
        """
        db_session = regular_api_client.app["db_session"]
        user = regular_api_client.app["test_authorized_user"]
        new_role = Role(name="test-role")
        db_session.add(new_role)
        db_session.flush()
        db_session.add(UserRole(user_id=user.id, role_id=new_role.id))
        db_session.commit()

        response = await regular_api_client.post("/test")
        assert response.status == 401
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {
            "message": "Your roles are out of date, please re-authenticate.",
            "hints": {}}

    async def test_invalid_token(self, regular_api_client):
        regular_api_client._session._default_headers.popall("authorization")
        response = await regular_api_client.post(
            "/test", headers={"Authorization": "bearer BAD"})
        assert response.status == 401
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {
            "message": "Invalid token.",
            "hints": {}}

    async def test_token_not_accepted(self, regular_api_client):
        """
        Missing tokens are invalid.
        """
        regular_api_client._session._default_headers.popall("authorization")
        response = await regular_api_client.post("/test")
        assert response.status == 401
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {
            "message": "Invalid token.",
            "hints": {}}


class TestCreateJWTToken:
    def test_simple_token(self):
        user = types.SimpleNamespace(
            id=24, name="Woo Hoo", email="woo@hoo",
            roles=[types.SimpleNamespace(name="admin"),
                   types.SimpleNamespace(name="user")])
        token = create_jwt_token(user, "secret")
        assert token == (
            "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzUxMiJ9."
            "eyJpZCI6MjQsIm5hbWUiOiJXb28gSG9vIiwiZW1haWwiOiJ"
            "3b29AaG9vIiwicm9sZXMiOlsiYWRtaW4iLCJ1c2VyIl19."
            "OWGEQzk6gb3S-I4zSNEf8O932InLFVEV6KcEL_h6b9J"
            "F3bQbGfcb8o3p-ZMbZFX98WFTMHOfeNJhJjzm4ItWMQ")

        header, payload, signature = token.split(".")
        header = header + "=" * (4 - len(header) % 4)
        payload = payload + "=" * (4 - len(payload) % 4)
        assert {"alg": "HS512", "typ": "JWT"} == json.loads(
            base64.b64decode(header))
        assert {"id": 24,
                "name": "Woo Hoo",
                "email": "woo@hoo",
                "roles": ["admin", "user"]} == json.loads(
                    base64.b64decode(payload))
