import datetime
import pytest

from aiohttp import web

from calorify_users.middleware import jwt_token_verification
from calorify_users.model.emailverification import (
    EmailVerification,
    datetime_24h_from_now,
)
from calorify_users.model.user import User
from calorify_users.view.emailverification import (
    confirm_reset_password,
    reset_password,
    verify_email_token,
)
from calorify_users.view.user import update_single_user


@pytest.fixture
def api_client(loop, aiohttp_client, testdb_session):
    app = web.Application(middlewares=[jwt_token_verification])
    app["db_session"] = testdb_session
    app["server_secret"] = "secret"
    app._app_config = {"product_name": "Product Name",
                       "base_url": "http://example/",
                       "allow_origin": "*",
                       "email": {
                           "from": "me@example",
                           "server": {
                               "host": "localhost"
                           }}}

    app.router.add_get('/test', verify_email_token)
    app.router.add_post('/pwtest', reset_password)
    app.router.add_post('/testlogin', confirm_reset_password)
    app.router.add_patch('/pwchange/{user_id}', update_single_user)
    yield loop.run_until_complete(
        aiohttp_client(app, headers={"accept": "application/json"}))


class TestEmailVerification:
    async def test_token_not_provided(self, api_client):
        response = await api_client.get("/test")
        assert response.status == 400, "Expected a 400 failure."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == {"message": "Invalid token.", "hints": {}}, (
            "Unexpected JSON content.")

    async def test_invalid_token(self, api_client):
        response = await api_client.get("/test?token=foobar")
        assert response.status == 400, "Expected a 400 failure."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == {"message": "Invalid token.", "hints": {}}, (
            "Unexpected JSON content.")

    async def test_token_missing(self, api_client):
        working_token = "0123456789abcdef0123456789abcdef"
        response = await api_client.get("/test?token={}".format(working_token))
        assert response.status == 400, "Expected a 400 failure."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == {
            "message": "Token not found or expired.", "hints": {}}, (
            "Unexpected JSON content.")

    async def test_token_matches(self, api_client):
        working_token = "0123456789abcdef0123456789abcdef"
        db_session = api_client.app["db_session"]
        user = User(name="Woo Hoo", password="password")
        db_session.add(user)
        db_session.flush()
        db_session.add(
            EmailVerification(
                user_id=user.id, email="woo@hoo", token=working_token,
                expires=datetime_24h_from_now()))
        db_session.commit()

        response = await api_client.get("/test?token={}".format(working_token))
        assert response.status == 200, "Should have succeeded."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == {"user_id": user.id}, "Unexpected JSON content."

        # User record is now updated, and the EmailVerification row is gone.
        updated_user = db_session.query(User).filter(
            User.id == user.id).one_or_none()
        assert updated_user.email == "woo@hoo"
        assert updated_user.email_verification is None

    async def test_with_url_callback(self, api_client):
        working_token = "0123456789abcdef0123456789abcdef"
        db_session = api_client.app["db_session"]
        user = User(name="Woo Hoo", password="password")
        db_session.add(user)
        db_session.flush()
        db_session.add(
            EmailVerification(
                user_id=user.id, email="woo@hoo", token=working_token,
                url_callback="https://somewhere/else",
                expires=datetime_24h_from_now()))
        db_session.commit()

        response = await api_client.get(
            "/test?token={}".format(working_token), allow_redirects=False)

        assert response.status == 302
        assert response.headers["Location"] == "https://somewhere/else"


class TestResetPassword:
    async def test_email_not_provided(self, api_client):
        response = await api_client.post("/pwtest")
        assert response.status == 400
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Email not specified.", "hints": {}}

    async def test_missing_user_noop(self, api_client):
        api_client.app._app_config["allow_origin"] = "https://example.com"

        response = await api_client.post(
            "/pwtest", data={"email": "not@there",
                             "url_callback": "https://example.com/?token="})
        assert response.status == 200
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"status": "success"}

    async def test_invalid_url_callback(self, api_client):
        api_client.app._app_config["allow_origin"] = "https://example.com"

        db_session = api_client.app["db_session"]
        db_session.add(User(email="woo@hoo", name="Woo Hoo"))
        db_session.commit()

        response = await api_client.post(
            "/pwtest", data={"email": "woo@hoo",
                             "url_callback": "http://not-example.com"})
        assert response.status == 400
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Invalid URL callback.", "hints": {}}

    async def test_works(self, api_client):
        api_client.app._app_config["allow_origin"] = "https://example.com"

        db_session = api_client.app["db_session"]
        db_session.add(User(email="woo@hoo", name="Woo Hoo"))
        db_session.commit()

        response = await api_client.post(
            "/pwtest", data={"email": "woo@hoo",
                             "url_callback": "https://example.com/?token="})
        assert response.status == 200
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"status": "success"}

        verification = db_session.query(EmailVerification).filter(
            EmailVerification.email == "woo@hoo").one_or_none()

        assert verification.is_password_reset


class TestConfirmResetPassword:
    async def test_token_not_provided(self, api_client):
        response = await api_client.post("/testlogin")
        assert response.status == 400
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Invalid token.", "hints": {}}

    async def test_token_missing(self, api_client):
        token = "1234567890abcdef1234567890abcdef"
        response = await api_client.post("/testlogin", data={"token": token})
        assert response.status == 400
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Token not found or expired.", "hints": {}}

    async def test_expired_token(self, api_client):
        working_token = "0123456789abcdef0123456789abcdef"
        db_session = api_client.app["db_session"]
        user = User(name="Woo Hoo", password="password", email="woo@hoo")
        db_session.add(user)
        db_session.flush()
        in_the_past = datetime.datetime.now() - datetime.timedelta(hours=1)
        db_session.add(
            EmailVerification(
                user_id=user.id, email="woo@hoo", token=working_token,
                url_callback="https://somewhere/else", is_password_reset=True,
                expires=in_the_past))
        db_session.commit()

        response = await api_client.post(
            "/testlogin", data={"token": working_token})

        assert response.status == 400
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Token not found or expired.", "hints": {}}

    async def test_token_not_for_password_reset(self, api_client):
        working_token = "0123456789abcdef0123456789abcdef"
        db_session = api_client.app["db_session"]
        user = User(name="Woo Hoo", password="password", email="woo@hoo")
        db_session.add(user)
        db_session.flush()
        db_session.add(
            EmailVerification(
                user_id=user.id, email="woo@hoo", token=working_token,
                url_callback="https://somewhere/else", is_password_reset=False,
                expires=datetime_24h_from_now()))
        db_session.commit()

        response = await api_client.post(
            "/testlogin", data={"token": working_token})

        assert response.status == 400
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Token not found or expired.", "hints": {}}

    async def test_works(self, api_client):
        working_token = "0123456789abcdef0123456789abcdef"
        db_session = api_client.app["db_session"]
        user = User(name="Woo Hoo", password="password", email="woo@hoo")
        db_session.add(user)
        db_session.flush()
        db_session.add(
            EmailVerification(
                user_id=user.id, email="woo@hoo", token=working_token,
                url_callback="https://somewhere/else", is_password_reset=True,
                expires=datetime_24h_from_now()))
        db_session.commit()

        response = await api_client.post(
            "/testlogin", data={"token": working_token})

        assert response.status == 200
        assert response.content_type == "application/json"
        data = await response.json()
        assert data["status"] == "success"
        assert data["token"] is not None

        # Changing password with this token now works.
        response = await api_client.patch(
            "/pwchange/{}".format(user.id), data={"password": "goodStrong1"},
            headers={"Authorization": "bearer " + data["token"]})
        assert response.status == 200
