import json
import pprint
import pytest
import types

from calorify_users.view.generic import (
    adapt_response,
    error_response,
)

request_variants = (
    types.SimpleNamespace(headers={}),
    types.SimpleNamespace(headers={"ACCEPT": "something/json+v1"}),
)


class TestErrorResponse:
    @pytest.mark.parametrize("request", request_variants,
                             ids=["no_header", "json_requested"])
    async def test_response_as_json(self, request):
        """If Accept header is missing we default to JSON."""

        message = "Test error message"
        hints = {"foo": "bar"}
        expected_data = {"message": message, "hints": hints}

        response = error_response(request, message, hints)

        assert response.status == 400, "Unexpected status code."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        assert json.loads(response.body) == expected_data, (
            "Unexpected JSON content.")

    async def test_not_json(self):
        """
        For any request explicitely asking for non-JSON data,
        we return only the error message as text/plain.
        """

        request = types.SimpleNamespace(headers={"ACCEPT": "text/html"})

        message = "Test error message"
        hints = {"foo": "bar"}

        response = error_response(request, message, hints)

        assert response.status == 400, "Unexpected status code."
        assert response.content_type == "text/plain", (
            "Unexpected content type.")
        assert response.text == message, "Unexpected text content."

    async def test_custom_status_code(self):
        """
        Return a custom status code on the response.
        """

        request = types.SimpleNamespace(headers={})
        response = error_response(request, "", {}, status=403)

        assert response.status == 403, "Unexpected status code."


class TestAdaptResponse:
    @pytest.mark.parametrize("request", request_variants,
                             ids=["no_header", "json_requested"])
    async def test_response_as_json(self, request):
        """If Accept header is missing or set to json, return json."""
        data = {"sample": [1, 2, 3]}
        response = adapt_response(request, data)

        assert response.status == 200
        assert response.content_type == "application/json"
        assert json.loads(response.body) == data

    async def test_not_json(self):
        """
        If Accept header asks for something other than json,
        return pretty-formatted data as text/plain.
        """
        request = types.SimpleNamespace(headers={"ACCEPT": "text/html"})
        data = {"sample": [1, 2, 3]}
        response = adapt_response(request, data)

        assert response.status == 200
        assert response.content_type == "text/plain"
        assert response.body == pprint.pformat(data).encode("utf-8")

    async def test_custom_status_code(self):
        """
        Return a custom status code on the response.
        """

        request = types.SimpleNamespace(headers={})
        response = adapt_response(request, "", status=403)

        assert response.status == 403
