import pytest

from aiohttp import web

from calorify_users.middleware import jwt_token_verification
from calorify_users.model.role import Role
from calorify_users.model.user import User
from calorify_users.model.userrole import UserRole
from calorify_users.view.authenticate import create_jwt_token
from calorify_users.view.role import (
    create_role,
    get_roles,
    remove_role,
)


def create_app_with_roles(testdb_session):
    """
    Returns a web.Application instance using JWT token verification
    and routing /test URLs to get_users/add_user.
    """
    app = web.Application(middlewares=[jwt_token_verification])
    app["db_session"] = testdb_session
    app["server_secret"] = "secret"
    app.router.add_get('/test', get_roles)
    app.router.add_post('/test', create_role)
    app.router.add_delete('/test/{name}', remove_role)
    return app


@pytest.fixture(params=["@admin", "@manager"])
def privileged_api_client(request, loop, aiohttp_client, testdb_session):
    app = create_app_with_roles(testdb_session)
    user = User(email="woo@hoo", name="Woo Hoo", password="pwd")
    role = Role(name=request.param)
    app["user_role"] = role.name
    testdb_session.add_all([user, role])
    testdb_session.flush()
    testdb_session.add(UserRole(user_id=user.id, role_id=role.id))
    testdb_session.commit()

    token = create_jwt_token(user, app["server_secret"])
    headers = {"Authorization": "bearer {}".format(token),
               "Accept": "application/json"}
    yield loop.run_until_complete(aiohttp_client(app, headers=headers))


@pytest.fixture
def anonymous_api_client(loop, aiohttp_client, testdb_session):
    app = create_app_with_roles(testdb_session)
    headers = {"Accept": "application/json"}
    yield loop.run_until_complete(aiohttp_client(app, headers=headers))


class TestGetRole:
    async def test_anonymous_denied(self, anonymous_api_client):
        response = await anonymous_api_client.get("/test")
        assert response.status == 401
        data = await response.json()
        assert data == {"message": "Unauthorized", "hints": {}}

    async def test_roles(self, privileged_api_client):
        app = privileged_api_client.app
        db_session = app["db_session"]
        db_session.add_all([Role(name="test1"), Role(name="test2")])
        db_session.flush()

        response = await privileged_api_client.get("/test")
        assert response.status == 200, "Expected a success."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == {"roles": [app["user_role"], "test1", "test2"]}, (
            "Unexpected JSON content.")


class TestCreateRole:
    async def test_new(self, privileged_api_client):
        response = await privileged_api_client.post(
            "/test", data={"name": "admin"})
        if privileged_api_client.app["user_role"] == "@manager":
            # Managers cannot create new roles.
            assert response.status == 401
            assert response.content_type == "application/json"
            data = await response.json()
            assert data == {"message": "Unauthorized", "hints": {}}
        else:
            assert response.status == 200, "Expected a success."
            assert response.content_type == "application/json", (
                "Unexpected content type.")
            data = await response.json()
            assert data == {"name": "admin"}, "Unexpected JSON content."

    async def test_invalid_role_name(self, privileged_api_client):
        if privileged_api_client.app["user_role"] == "@manager":
            pytest.skip("Managers cannot create new roles, skip this test.")
        response = await privileged_api_client.post(
            "/test", data={"name": "a"})
        assert response.status == 400, "Expected 400 Bad Request failure."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == {
            "message": "Role name must be at least 2 characters long.",
            "hints": {}}, "Unexpected JSON content."

    async def test_duplicate_role(self, privileged_api_client):
        if privileged_api_client.app["user_role"] == "@manager":
            pytest.skip("Managers cannot create new roles, skip this test.")
        db_session = privileged_api_client.app["db_session"]
        db_session.add(Role(name="admin"))
        db_session.flush()

        response = await privileged_api_client.post(
            "/test", data={"name": "admin"})
        assert response.status == 400, "Expected 400 Bad Request failure."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == {
            "message": "Role 'admin' already exists.",
            "hints": {}}, "Unexpected JSON content."


class TestRemoveRole:
    async def test_remove_missing(self, privileged_api_client):
        response = await privileged_api_client.delete("/test/unknown")
        if privileged_api_client.app["user_role"] == "@manager":
            # Managers cannot delete roles.
            assert response.status == 401
            assert response.content_type == "application/json"
            data = await response.json()
            assert data == {"message": "Unauthorized", "hints": {}}
        else:
            assert response.status == 404
            assert response.content_type == "application/json"
            data = await response.json()
            assert data == {"message": "Role 'unknown' not found.",
                            "hints": {}}

    async def test_works(self, privileged_api_client):
        if privileged_api_client.app["user_role"] == "@manager":
            pytest.skip("Managers cannot create new roles, skip this test.")
        db_session = privileged_api_client.app["db_session"]
        db_session.add(Role(name="testrole"))
        db_session.flush()

        response = await privileged_api_client.delete("/test/testrole")
        assert response.status == 200
        assert response.content_type == "application/json"

        roles = [role.name for role in db_session.query(Role).all()]
        assert "testrole" not in roles
        assert roles == ['@admin']
