import pytest

from aiohttp import web

from calorify_users.middleware import jwt_token_verification
from calorify_users.model.role import Role
from calorify_users.model.user import User
from calorify_users.model.userrole import UserRole
from calorify_users.view.authenticate import create_jwt_token
from calorify_users.view.user import (
    add_user,
    get_single_user,
    get_users,
    remove_single_user,
    update_single_user,
    validate_new_user_data,
    validate_password_strength,
    validate_url_from_allowed_origin,
)


def create_users_app(testdb_session):
    """
    Returns a web.Application instance using JWT token verification
    and routing /test URLs to get_users/add_user.
    """
    app = web.Application(middlewares=[jwt_token_verification])
    app["db_session"] = testdb_session
    app["server_secret"] = "secret"
    app._app_config = {}
    app.router.add_get('/test', get_users)
    app.router.add_post('/test', add_user)
    app.router.add_get('/test/{user_id}', get_single_user)
    app.router.add_patch('/test/{user_id}', update_single_user)
    app.router.add_delete('/test/{user_id}', remove_single_user)
    return app


@pytest.fixture(params=["@admin", "@manager"])
def privileged_api_client(request, loop, aiohttp_client, testdb_session):
    app = create_users_app(testdb_session)
    user = User(email="woo@hoo", name="Woo Hoo", password="pwd")
    role = Role(name=request.param)
    testdb_session.add_all([user, role])
    testdb_session.flush()
    testdb_session.add(UserRole(user_id=user.id, role_id=role.id))
    testdb_session.commit()

    token = create_jwt_token(user, app["server_secret"])
    headers = {"Authorization": "bearer {}".format(token),
               "Accept": "application/json"}
    yield loop.run_until_complete(aiohttp_client(app, headers=headers))


@pytest.fixture
def regular_api_client(loop, aiohttp_client, testdb_session):
    app = create_users_app(testdb_session)
    user = User(email="woo@hoo", name="Woo Hoo", password="pwd")
    testdb_session.add(user)
    testdb_session.commit()
    token = create_jwt_token(user, app["server_secret"])
    headers = {"Authorization": "bearer {}".format(token),
               "Accept": "application/json"}
    app["test_authorized_user"] = user
    yield loop.run_until_complete(aiohttp_client(app, headers=headers))


@pytest.fixture
def anonymous_api_client(loop, aiohttp_client, testdb_session):
    app = create_users_app(testdb_session)
    headers = {"Accept": "application/json"}
    yield loop.run_until_complete(aiohttp_client(app, headers=headers))


class TestGetUsers:
    async def test_get_users_one(self, privileged_api_client):
        """
        Listing users works when logged in with roles "@admin" or "@manager".
        """
        response = await privileged_api_client.get("/test")
        assert response.status == 200, "Failed call to get_users."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == [{"id": 1, "name": "Woo Hoo", "email": "woo@hoo"}], (
            "Unexpected JSON content.")

    async def test_get_users_multiple(self, privileged_api_client):
        """
        Listing users works when logged in with roles "@admin" or "@manager".
        """
        db_session = privileged_api_client.app["db_session"]
        db_session.add(User(email="foo@bar", name="Foo Bar"))

        response = await privileged_api_client.get("/test")
        assert response.status == 200, "Failed call to get_users."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == [
            {"id": 1, "name": "Woo Hoo", "email": "woo@hoo"},
            {"id": 2, "name": "Foo Bar", "email": "foo@bar"},
        ], ("Unexpected JSON content.")

    async def test_get_users_fails_anonymously(self, anonymous_api_client):
        """
        Listing users fails with 401 error for anonymous clients.
        """
        response = await anonymous_api_client.get("/test")
        assert response.status == 401
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Unauthorized", "hints": {}}


class TestAddUser:
    async def test_create_user(self, anonymous_api_client):
        response = await anonymous_api_client.post(
            "/test",
            data={"email": "woo@hoo", "name": "Woo Hoo",
                  "password": "1StrongPassword"})
        assert response.status == 200, "Failed call to add_user."
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        assert data == {"user_id": 1}, "Unexpected JSON content."

        db_session = anonymous_api_client.app["db_session"]
        new_user = db_session.query(User).one_or_none()
        assert new_user.email is None
        assert new_user.email_verification.email == "woo@hoo"
        assert not new_user.email_verification.is_email_change

    async def test_create_user_with_callback(self, anonymous_api_client):
        config = {"allow_origin": "https://example.com:1234"}
        anonymous_api_client.app._app_config = config

        response = await anonymous_api_client.post(
            "/test",
            data={"email": "woo@hoo", "name": "Woo Hoo",
                  "password": "1StrongPassword",
                  "url_callback": "https://example.com:1234/trusted"})
        assert response.status == 200
        data = await response.json()
        assert data == {"user_id": 1}, "Unexpected JSON content."

    async def test_create_duplicate_user(self, anonymous_api_client):
        db_session = anonymous_api_client.app["db_session"]
        db_session.add(User(email="woo@hoo", name="Woo Hoo"))
        db_session.flush()

        response = await anonymous_api_client.post(
            "/test",
            data={"email": "woo@hoo", "name": "Woo Hoo",
                  "password": "1Password"},
            headers={"accept": "application/json"})
        assert response.status == 400, (
            "Creating a user with duplicate email should have failed.")
        assert response.content_type == "application/json", (
            "Unexpected content type.")
        data = await response.json()
        expected_data = {
            "message": "User with address woo@hoo already registered.",
            "hints": {
                "fields": {
                    "email": ["This email is already registered."],
                }
            }
        }
        assert data == expected_data, "Unexpected JSON content."


class TestGetSingleUser:
    async def test_invalid_user_id(self, regular_api_client):
        response = await regular_api_client.get("/test/not-an-id")

        assert response.status == 400
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Invalid user ID 'not-an-id'.", "hints": {}}

    async def test_anonymous(self, anonymous_api_client):
        db_session = anonymous_api_client.app["db_session"]
        user = User(email="woo@hoo", name="Woo Hoo")
        db_session.add(user)
        db_session.flush()

        response = await anonymous_api_client.get("/test/{}".format(user.id))

        assert response.status == 401
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Unauthorized", "hints": {}}

    async def test_get_own_data(self, regular_api_client):
        user = regular_api_client.app["test_authorized_user"]

        response = await regular_api_client.get("/test/{}".format(user.id))

        assert response.status == 200
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"name": user.name, "email": user.email}

    async def test_get_elses_data(self, regular_api_client):
        db_session = regular_api_client.app["db_session"]
        user = User(email="foo@bar", name="Foo Bar")
        db_session.add(user)
        db_session.flush()

        response = await regular_api_client.get("/test/{}".format(user.id))

        assert response.status == 401
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Unauthorized", "hints": {}}

    async def test_privileged_get_elses_data(self, privileged_api_client):
        db_session = privileged_api_client.app["db_session"]
        user = User(email="someone@new", name="Someone New")
        db_session.add(user)
        db_session.flush()

        response = await privileged_api_client.get("/test/{}".format(user.id))

        assert response.status == 200
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"name": user.name, "email": user.email}


class TestRemoveUser:
    async def test_anonymous(self, anonymous_api_client):
        db_session = anonymous_api_client.app["db_session"]
        user = User(email="woo@hoo", name="Woo Hoo")
        db_session.add(user)
        db_session.flush()

        response = await anonymous_api_client.delete(
            "/test/{}".format(user.id))

        assert response.status == 401
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Unauthorized", "hints": {}}

        updated_user = db_session.query(User).filter(
            User.id == user.id).one_or_none()
        assert updated_user is not None

    async def test_delete_myself(self, regular_api_client):
        user = regular_api_client.app["test_authorized_user"]

        response = await regular_api_client.delete(
            "/test/{}".format(user.id))

        assert response.status == 200
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"name": user.name, "email": user.email}

        db_session = regular_api_client.app["db_session"]
        updated_user = db_session.query(User).filter(
            User.id == user.id).one_or_none()
        assert updated_user is None

    async def test_delete_someone_else(self, regular_api_client):
        db_session = regular_api_client.app["db_session"]
        user = User(email="foo@bar", name="Foo Bar")
        db_session.add(user)
        db_session.flush()

        response = await regular_api_client.delete(
            "/test/{}".format(user.id))

        assert response.status == 401
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Unauthorized", "hints": {}}

        updated_user = db_session.query(User).filter(
            User.id == user.id).one_or_none()
        assert updated_user is not None

    async def test_privileged_delete_someone(self, privileged_api_client):
        db_session = privileged_api_client.app["db_session"]
        user = User(email="someone@new", name="Someone New")
        db_session.add(user)
        db_session.flush()

        response = await privileged_api_client.delete(
            "/test/{}".format(user.id))

        assert response.status == 200

        updated_user = db_session.query(User).filter(
            User.id == user.id).one_or_none()
        assert updated_user is None


class TestUpdateUser:
    async def test_anonymous(self, anonymous_api_client):
        db_session = anonymous_api_client.app["db_session"]
        user = User(email="woo@hoo", name="Woo Hoo")
        db_session.add(user)
        db_session.flush()

        response = await anonymous_api_client.patch(
            "/test/{}".format(user.id), data={"name": "New Name"})

        assert response.status == 401
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Unauthorized", "hints": {}}

    async def test_update_own_data(self, regular_api_client):
        user = regular_api_client.app["test_authorized_user"]

        response = await regular_api_client.patch(
            "/test/{}".format(user.id),
            data={"name": "New Name", "email": user.email})

        assert response.status == 200
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"name": "New Name", "email": user.email}

    async def test_update_elses_data(self, regular_api_client):
        db_session = regular_api_client.app["db_session"]
        user = User(email="foo@bar", name="Foo Bar")
        db_session.add(user)
        db_session.flush()

        response = await regular_api_client.patch(
            "/test/{}".format(user.id),
            data={"name": "New Name", "email": user.email})

        assert response.status == 401
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"message": "Unauthorized", "hints": {}}

    async def test_privileged_update_elses_data(self, privileged_api_client):
        db_session = privileged_api_client.app["db_session"]
        user = User(email="someone@new", name="Someone New")
        db_session.add(user)
        db_session.flush()

        response = await privileged_api_client.patch(
            "/test/{}".format(user.id),
            data={"name": "New Name", "email": user.email})

        assert response.status == 200
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"name": "New Name", "email": user.email}

    async def test_update_only_name(self, regular_api_client):
        user = regular_api_client.app["test_authorized_user"]

        response = await regular_api_client.patch(
            "/test/{}".format(user.id),
            data={"name": "New Name"})

        assert response.status == 200
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {"name": "New Name", "email": user.email}

    async def test_update_name_empty_error(self, regular_api_client):
        user = regular_api_client.app["test_authorized_user"]

        response = await regular_api_client.patch(
            "/test/{}".format(user.id), data={"name": ""})

        assert response.status == 400
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {
            "message": "Empty name provided.",
            "hints": {"fields": {
                "name": ["New name must not be an empty string."]}}}

    async def test_update_only_password(self, regular_api_client):
        user = regular_api_client.app["test_authorized_user"]
        old_hash = user.password_hash

        response = await regular_api_client.patch(
            "/test/{}".format(user.id),
            data={"password": "NewPassword2"})

        assert response.status == 200
        db_session = regular_api_client.app["db_session"]
        updated_user = db_session.query(User).filter(
            User.id == user.id).one_or_none()
        assert updated_user.password_hash != old_hash

    async def test_empty_password_ignored(self, regular_api_client):
        user = regular_api_client.app["test_authorized_user"]
        old_hash = user.password_hash

        response = await regular_api_client.patch(
            "/test/{}".format(user.id),
            data={"password": "   "})

        assert response.status == 200
        db_session = regular_api_client.app["db_session"]
        updated_user = db_session.query(User).filter(
            User.id == user.id).one_or_none()
        assert updated_user.password_hash == old_hash

    async def test_update_password_weak_error(self, regular_api_client):
        config = {"password_constraints":
                  {"minimum_length": 5,
                   "minimum_character_classes": 1}}
        regular_api_client.app._app_config = config

        user = regular_api_client.app["test_authorized_user"]

        response = await regular_api_client.patch(
            "/test/{}".format(user.id),
            data={"password": "a"})

        assert response.status == 400
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {
            "message": "Password not strong enough.",
            "hints": {"fields": {
                "password": ["Password must be at least 5 characters long."]}}}

    async def test_update_only_email(self, regular_api_client):
        user = regular_api_client.app["test_authorized_user"]
        old_email = user.email

        response = await regular_api_client.patch(
            "/test/{}".format(user.id),
            data={"email": "new@mail"})

        assert response.status == 200
        db_session = regular_api_client.app["db_session"]
        updated_user = db_session.query(User).filter(
            User.id == user.id).one_or_none()
        assert updated_user.email == old_email
        assert updated_user.email != "new@mail"
        assert updated_user.email_verification.email == "new@mail"
        assert updated_user.email_verification.is_email_change

    async def test_update_email_already_used(self, regular_api_client):
        db_session = regular_api_client.app["db_session"]
        db_session.add(User(name="Foo Bar", email="new@mail"))
        db_session.flush()

        user = regular_api_client.app["test_authorized_user"]

        response = await regular_api_client.patch(
            "/test/{}".format(user.id),
            data={"email": "new@mail"})

        assert response.status == 400
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {
            "message": "User with address new@mail already registered.",
            "hints": {"fields": {
                "email": ["This email is already registered."]}}}

    async def test_update_email_empty_error(self, regular_api_client):
        user = regular_api_client.app["test_authorized_user"]

        response = await regular_api_client.patch(
            "/test/{}".format(user.id), data={"email": ""})

        assert response.status == 400
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {
            "message": "Empty email provided.",
            "hints": {"fields": {
                "email": ["New email address must not be an empty string."]}}}

    async def test_update_email_invalid_error(self, regular_api_client):
        user = regular_api_client.app["test_authorized_user"]

        response = await regular_api_client.patch(
            "/test/{}".format(user.id), data={"email": ".foo@bar"})

        assert response.status == 400
        assert response.content_type == "application/json"
        data = await response.json()
        assert data == {
            "message": "Invalid e-mail address.",
            "hints": {"fields": {
                "email": ["Invalid e-mail address."]}}}


class TestValidateNewUserData:
    sample_user_data = {
        "email": "woo@hoo",
        "name": "Woo Hoo",
        "password": "1StrongPassword",
        "url_callback": None,
    }

    def test_validation_passes(self, testdb_session):
        test_data = self.sample_user_data.copy()
        result_data, error, hints = validate_new_user_data(
            test_data, {}, testdb_session)
        assert (test_data == result_data and error is None), (
            "Validation unexpectedly failed.")

    def test_missing_name(self, testdb_session):
        test_data = self.sample_user_data.copy()
        del test_data["name"]
        result_data, error, hints = validate_new_user_data(
            test_data, {}, testdb_session)
        assert (
            result_data["name"] == "" and
            error == "One or more required fields are missing." and
            hints == {
                "fields": {"name": ["User's name must be provided."]}
            }), ("Validation unexpectedly passed.")

    def test_empty_name(self, testdb_session):
        test_data = self.sample_user_data.copy()
        test_data["name"] = ""
        result_data, error, hints = validate_new_user_data(
            test_data, {}, testdb_session)
        assert (
            result_data["name"] == "" and
            error == "One or more required fields are missing." and
            hints == {
                "fields": {"name": ["User's name must be provided."]}
            }), ("Validation unexpectedly passed.")

    def test_missing_email(self, testdb_session):
        test_data = self.sample_user_data.copy()
        del test_data["email"]
        result_data, error, hints = validate_new_user_data(
            test_data, {}, testdb_session)
        assert (
            result_data["email"] == "" and
            error == "One or more required fields are missing." and
            hints == {
                "fields": {"email": ["User's email must be provided."]}
            }), ("Validation unexpectedly passed.")

    def test_empty_email(self, testdb_session):
        test_data = self.sample_user_data.copy()
        test_data["email"] = ""
        result_data, error, hints = validate_new_user_data(
            test_data, {}, testdb_session)
        assert (
            result_data["email"] == "" and
            error == "One or more required fields are missing." and
            hints == {
                "fields": {"email": ["User's email must be provided."]}
            }), ("Validation unexpectedly passed.")

    def test_missing_password(self, testdb_session):
        test_data = self.sample_user_data.copy()
        del test_data["password"]
        result_data, error, hints = validate_new_user_data(
            test_data, {}, testdb_session)
        assert (
            result_data["password"] == "" and
            error == "One or more required fields are missing." and
            hints == {
                "fields": {"password": ["User's password must be provided."]}
            }), ("Validation unexpectedly passed.")

    def test_empty_password(self, testdb_session):
        test_data = self.sample_user_data.copy()
        test_data["password"] = ""
        result_data, error, hints = validate_new_user_data(
            test_data, {}, testdb_session)
        assert (
            result_data["password"] == "" and
            error == "One or more required fields are missing." and
            hints == {
                "fields": {"password": ["User's password must be provided."]}
            }), ("Validation unexpectedly passed.")

    def test_weak_password(self, testdb_session):
        test_data = self.sample_user_data.copy()
        config = {"password_constraints": {
            "minimum_length": 5, "minimum_character_classes": 1}}
        test_data["password"] = "weak"
        result_data, error, hints = validate_new_user_data(
            test_data, config, testdb_session)
        assert (
            result_data["password"] == "weak" and
            error == "Password is not strong enough." and
            hints == {
                "fields": {
                    "password": [
                        "Password must be at least 5 characters long."]}
            }), ("Validation unexpectedly passed.")

    def test_missing_url_callback(self, testdb_session):
        test_data = self.sample_user_data.copy()
        del test_data["url_callback"]
        result_data, error, hints = validate_new_user_data(
            test_data, {}, testdb_session)
        assert (result_data["url_callback"] is None and
                error is None and hints == {"fields": {}})

    def test_empty_url_callback(self, testdb_session):
        test_data = self.sample_user_data.copy()
        test_data["url_callback"] = ""
        result_data, error, hints = validate_new_user_data(
            test_data, {}, testdb_session)
        assert (result_data["url_callback"] is None and
                error is None and hints == {"fields": {}})

    def test_bad_url_callback(self, testdb_session):
        config = {"allow_origin": "https://example.com:1234"}

        test_data = self.sample_user_data.copy()
        test_data["url_callback"] = "http://example.com/untrusted"
        result_data, error, hints = validate_new_user_data(
            test_data, config, testdb_session)
        assert (result_data["url_callback"] is None and
                error == "Callback URL is not allowed." and
                hints == {
                    "fields": {
                        "url_callback": [
                            "Callback URL must match allowed origin."]}
                })


class TestValidatePasswordStrength:
    def test_defaults_too_short(self):
        """Defaults ask for a password at least 8 characters long."""
        expected_error = "Password must be at least 8 characters long."
        assert validate_password_strength("toofew1", {}) == expected_error

    def test_defaults_too_weak(self):
        """Defaults ask for a password with at least 3 types of characters."""
        expected_error = (
            "Password must use at least 3 classes of characters "
            "(eg. uppercase, lowercase, digits, special symbols.")
        assert validate_password_strength("tooweak1", {}) == expected_error

    def test_defaults_passes(self):
        """Defaults ask for a password with at least 3 types of characters."""
        assert validate_password_strength("1StrongPassword", {}) is None

    def test_uses_config(self):
        constraints = {"minimum_length": 5, "minimum_character_classes": 1}
        assert validate_password_strength("passes", constraints) is None


class TestValidateUrlFromAllowedOrigin:
    def test_no_callback_is_fine(self):
        assert validate_url_from_allowed_origin(None, "https://foo")

    def test_empty_callback_is_fine(self):
        assert validate_url_from_allowed_origin("  ", "https://foo")

    def test_domain_matching_url_is_fine(self):
        assert validate_url_from_allowed_origin(
            "https://foo/trusted", "https://foo")

    def test_any_origin_is_ok(self):
        assert validate_url_from_allowed_origin(
            "http://somewhere/trusted", "*")

    def test_explicit_port_vs_no_port_is_bad(self):
        assert not validate_url_from_allowed_origin(
            "https://foo:443/trusted", "https://foo")

    def test_bad_port_is_bad(self):
        assert not validate_url_from_allowed_origin(
            "https://foo:8000/untrusted", "https://foo")

    def test_bad_scheme_is_bad(self):
        assert not validate_url_from_allowed_origin(
            "http://foo/untrusted", "https://foo")

    def test_other_domain_is_bad(self):
        assert not validate_url_from_allowed_origin(
            "https://somewhere/untrusted", "https://foo")
