import unicodedata

from sqlalchemy.sql.expression import null
from urllib.parse import urljoin

from calorify_users.helpers import validate_email
from calorify_users.model.emailverification import (
    create_or_update_email_verification,
)
from calorify_users.model.user import User
from calorify_users.permissions import require_role
from calorify_users.tasks.emailverification import send_email_verification
from calorify_users.view.generic import (
    adapt_response,
    error_response,
)


@require_role("@admin", "@manager")
async def get_users(request):
    """
    ---
    description: List all users managed by this service.
    tags:
    - Users
    produces:
    - application/json
    responses:
        "200":
            description: return a collection of all users.
    """
    db_session = request.app["db_session"]
    users = db_session.query(User).filter(
        User.email != null()).order_by(User.id).all()
    users_data = [
        {"id": user.id,
         "name": user.name,
         "email": user.email}
        for user in users]
    return adapt_response(request, users_data)


async def add_user(request):
    """
    ---
    description:
        Creates a new user with their email address marked as unverified.
    tags:
    - Users
    produces:
    - application/json
    responses:
        "200":
            description: return a new user record along with Link relation.
    """
    user_data = await request.post()
    db_session = request.app["db_session"]

    user_data, error_message, error_hints = validate_new_user_data(
        user_data, request.app._app_config, db_session)

    if error_message is not None:
        return error_response(request, error=error_message, hints=error_hints)
    new_user = User(name=user_data["name"], password=user_data["password"])
    db_session.add(new_user)
    db_session.flush()
    verification = create_or_update_email_verification(
        db_session, new_user.id, user_data["email"], user_data["url_callback"])
    db_session.commit()
    request.loop.create_task(
        send_email_verification(
            verification, db_session, request.app._app_config))
    return adapt_response(request, {"user_id": new_user.id})


def can_access_user_data(request, request_match_field="user_id"):
    """
    Check if logged in user (in `request`) can access user's data.
    User ID is taken from the `request_match_field`-named field in the URL.

    Return a pair indicating user given access to and any error response.
    If there is an error, user will be None.
    """
    db_session = request.app["db_session"]
    try:
        user_id = int(request.match_info.get("user_id"))
    except ValueError:
        return None, error_response(
            request, "Invalid user ID '{}'.".format(
                request.match_info.get("user_id")), {})

    user = db_session.query(User).filter(
        User.email != null(), User.id == user_id).one_or_none()

    authorized_roles = set(request.get("authorized_roles", set()))
    has_privileged_role = bool(
        authorized_roles.intersection({"@admin", "@manager"}))
    authorized_user = request.get("authorized_user")

    if not has_privileged_role and user != authorized_user:
        return None, error_response(request, "Unauthorized", {}, status=401)
    return user, None


async def get_single_user(request):
    """
    ---
    description: Get details for a single user.
    tags:
    - Users
    produces:
    - application/json
    responses:
        "200":
            description: return a JSON object representing the user resource.
    """
    user, error_response = can_access_user_data(request)
    if error_response is not None:
        return error_response
    user_data = {"name": user.name, "email": user.email}
    return adapt_response(request, user_data)


async def remove_single_user(request):
    """
    ---
    description: Remove a user from the database.
    tags:
    - Users
    produces:
    - application/json
    responses:
        "200":
            description: return a JSON object representing the user resource.
    """
    user, error_response = can_access_user_data(request)
    if error_response is not None:
        return error_response
    user_data = {"name": user.name, "email": user.email}
    request.app["db_session"].delete(user)
    request.app["db_session"].commit()
    return adapt_response(request, user_data)


async def update_single_user(request):
    """
    ---
    description: Update details for a particular user. No field is required.
    tags:
    - Users
    produces:
    - application/json
    responses:
        "200":
            description: return a JSON object representing the user resource.
    """
    user, error_result = can_access_user_data(request)
    if error_result is not None:
        return error_result

    user_data = await request.post()
    db_session = request.app["db_session"]

    error_messages = []
    hints = {"fields": {}}
    name = user_data.get("name", None)
    if user_data.get("name") == "":
        # If name was passed in, it must not be an empty string.
        hints["fields"]["name"] = ["New name must not be an empty string."]
        error_messages.append("Empty name provided.")
    else:
        user.name = name

    email = user_data.get("email")
    if email == "":
        # If email was passed in, it must not be an empty string.
        hints["fields"]["email"] = [
            "New email address must not be an empty string."]
        error_messages.append("Empty email provided.")
    elif email is not None and email != user.email:
        # Attempt to update the email address when it is passed in and
        # if it's different from what we currently have.
        try:
            email = validate_email(user_data.get("email"))
        except AssertionError as error:
            error_messages.append("Invalid e-mail address.")
            hints["fields"].setdefault("email", []).append(str(error))
        else:
            email_already_used = db_session.query(User).filter(
                User.email == email).one_or_none()
            if email_already_used:
                hints["fields"].setdefault("email", []).append(
                    "This email is already registered.")
                error_messages.append(
                    "User with address {} already registered.".format(email))
            else:
                url_callback = validate_url_callback(
                    user_data.get("url_callback"), request.app._app_config,
                    error_messages, hints)
                if not error_messages:
                    # Valid emails still require verification.
                    verification = create_or_update_email_verification(
                        db_session, user.id, email, url_callback, True)
                    db_session.commit()
                    request.loop.create_task(
                        send_email_verification(
                            verification, db_session, request.app._app_config))

    password = user_data.get("password")
    if password is not None and password.strip() != "":
        error = validate_password_strength(
            password, request.app._app_config.get("password_constraints", {}))
        if error is None:
            user.password = password
        else:
            hints["fields"]["password"] = [error]
            error_messages.append("Password not strong enough.")

    db_session.commit()

    if error_messages:
        return error_response(request, " ".join(error_messages), hints)

    user_data = {"name": user.name,
                 "email": user.email}
    return adapt_response(request, user_data)


def validate_new_user_data(user_data, config, db_session):
    """Validates submitted user_data.

    Ensures that all required fields are present and that email address is not
    already used by an existing user if `new_user` is True.

    Returns a tuple (user_data, error_message, error_hints) where error_message
    is None if there are no errors.
    """

    error_messages = []
    error_hints = {"fields": {}}

    # Ensure all required fields are present.
    if not(user_data.get("name") and
           user_data.get("email") and
           user_data.get("password")):
        error_messages.append("One or more required fields are missing.")
        if not user_data.get("name"):
            error_hints["fields"]["name"] = ["User's name must be provided."]
        if not user_data.get("email"):
            error_hints["fields"]["email"] = ["User's email must be provided."]
        if not user_data.get("password"):
            error_hints["fields"]["password"] = [
                "User's password must be provided."]
    full_name = user_data.get("name", "")
    email = user_data.get("email", "")
    password = user_data.get("password", "")
    url_callback = user_data.get("url_callback")

    url_callback = validate_url_callback(
        url_callback, config, error_messages, error_hints)

    if password != "":
        password_strength = validate_password_strength(
            password, config.get("password_constraints", {}))
        if password_strength is not None:
            error_hints["fields"].setdefault("password", []).append(
                password_strength)
            error_messages.append("Password is not strong enough.")

    if email != "" and db_session.query(User).filter(
            User.email == email).one_or_none():
        email_errors = error_hints["fields"].setdefault("email", [])
        email_errors.append("This email is already registered.")
        error_messages.append(
            "User with address {} already registered.".format(email))

    return ({"name": full_name, "email": email, "password": password,
             "url_callback": url_callback},
            " ".join(error_messages) if error_messages else None,
            error_hints)


def validate_url_callback(url_callback, config, error_messages, error_hints):
    if url_callback is None:
        url_callback = ""
    else:
        url_callback = url_callback.strip()

    callback_valid = validate_url_from_allowed_origin(
        url_callback, config.get("allow_origin", "*"))
    if not callback_valid:
        error_messages.append("Callback URL is not allowed.")
        error_hints["fields"]["url_callback"] = [
            "Callback URL must match allowed origin."]
        url_callback = None
    else:
        if url_callback == "":
            url_callback = None
    return url_callback


def validate_url_from_allowed_origin(url, origin):
    """Return `url` if it is allowed from `origin`, otherwise None."""
    if url is None or url.strip() == "":
        return True

    base_url = urljoin(url, "/")
    base_url_without_slash = base_url[:-1] if base_url[-1] == "/" else base_url
    if origin == "*" or origin in (base_url, base_url_without_slash):
        return True
    return False


def validate_password_strength(password, password_constraints):
    """Ensure password is strong enough.

    Returns an error message or None.
    """
    minimum_length = password_constraints.get("minimum_length", 8)
    if len(password) < minimum_length:
        return "Password must be at least {} characters long.".format(
            minimum_length)

    character_classes = set()
    for char in password:
        character_classes.add(unicodedata.category(char))
    required_classes = password_constraints.get("minimum_character_classes", 3)
    if len(character_classes) < required_classes:
        return (
            "Password must use at least {} classes of characters "
            "(eg. uppercase, lowercase, digits, special symbols.").format(
                required_classes)

    return None
